import { Chain, ChainToPlatform, Network, Platform } from "@wormhole-foundation/sdk-base";
import { AccountAddress, ChainAddress, ChainContext, Contracts, NativeAddress, PayloadDiscriminator, PayloadLiteral, PlatformContext, PlatformUtils, TokenAddress, TokenId, TxHash, UniversalAddress, WormholeMessageId, deserialize } from "@wormhole-foundation/sdk-definitions";
import { WormholeConfig } from "./types";
import { CircleTransfer } from "./protocols/cctpTransfer";
import { TokenTransfer } from "./protocols/tokenTransfer";
import { TransactionStatus } from "./whscan-api";
export declare class Wormhole<N extends Network> {
    protected readonly _network: N;
    protected _platforms: Map<Platform, PlatformContext<N, Platform>>;
    protected _chains: Map<Chain, ChainContext<N, Platform, Chain>>;
    readonly config: WormholeConfig;
    constructor(network: N, platforms: PlatformUtils<N, any>[], config?: WormholeConfig);
    get network(): N;
    /**
     * Creates a CircleTransfer object to move Native USDC from one chain to another
     * @param amount the amount to transfer
     * @param from the address to transfer from
     * @param to the address to transfer to
     * @param automatic whether to use automatic delivery
     * @param payload the payload to send with the transfer
     * @param nativeGas the amount of native gas to send with the transfer
     * @returns the CircleTransfer object
     * @throws Errors if the chain or protocol is not supported
     */
    circleTransfer(amount: bigint, from: ChainAddress, to: ChainAddress, automatic: boolean, payload?: Uint8Array, nativeGas?: bigint): Promise<CircleTransfer<N>>;
    /**
     * Creates a TokenTransfer object to move a token from one chain to another
     * @param token the token to transfer
     * @param amount the amount to transfer
     * @param from the address to transfer from
     * @param to the address to transfer to
     * @param automatic whether to use automatic delivery
     * @param payload the payload to send with the transfer
     * @param nativeGas the amount of native gas to send with the transfer
     * @returns the TokenTransfer object
     * @throws Errors if the chain or protocol is not supported
     */
    tokenTransfer(token: TokenId | "native", amount: bigint, from: ChainAddress, to: ChainAddress, automatic: boolean, payload?: Uint8Array, nativeGas?: bigint): Promise<TokenTransfer<N>>;
    /**
     * Gets the contract addresses for a given chain
     * @param chain the chain name or chain id
     * @returns the contract addresses
     */
    getContracts(chain: Chain): Contracts | undefined;
    /**
     * Returns the platform object, i.e. the class with platform-specific logic and methods
     * @param chain the chain name or platform name
     * @returns the platform context class
     * @throws Errors if platform is not found
     */
    getPlatform<P extends Platform>(platformName: P): PlatformContext<N, P>;
    /**
     * Returns the chain "context", i.e. the class with chain-specific logic and methods
     * @param chain the chain name
     * @returns the chain context class
     * @throws Errors if context is not found
     */
    getChain<C extends Chain, P extends ChainToPlatform<C>>(chain: C): ChainContext<N, P, C>;
    /**
     * Gets the TokenId for a token representation on any chain
     *  These are the Wormhole wrapped token addresses, not necessarily
     *  the cannonical version of that token
     *
     * @param chain The chain name or id to get the wrapped token address
     * @param tokenId The Token ID (chain/address) of the original token
     * @returns The TokenId on the given chain, null if it does not exist
     * @throws Errors if the chain is not supported or the token does not exist
     */
    getWrappedAsset<C extends Chain>(chain: C, token: TokenId<Chain>): Promise<TokenId<C>>;
    /**
     * Gets the number of decimals for a token on a given chain
     *
     * @param chain The chain name or id of the token/representation
     * @param token The token address
     * @returns The number of decimals
     */
    getDecimals<C extends Chain>(chain: C, token: TokenAddress<C>): Promise<bigint>;
    /**
     * Fetches the balance of a given token for a wallet
     *
     * @param walletAddress The wallet address
     * @param tokenId The token ID (its home chain and address on the home chain)
     * @param chain The chain name or id
     * @returns The token balance of the wormhole asset as a BigNumber
     */
    getBalance<C extends Chain>(chain: C, token: TokenAddress<C>, walletAddress: string): Promise<bigint | null>;
    /**
     * Gets the associated token account for chains that require it (only Solana currently).
     * @param token the TokenId of the token to get the token account for
     * @param recipient the address of the primary account that may require a separate token account
     * @returns
     */
    getTokenAccount<SC extends Chain, RC extends Chain>(recipient: ChainAddress<RC>, token: TokenId<SC>): Promise<ChainAddress<RC>>;
    /**
     * Gets the Raw VAA Bytes from the API or Guardian RPC, finality must be met before the VAA will be available.
     * @param chain The chain name
     * @param emitter The emitter address
     * @param sequence The sequence number
     * @param retries The number of times to retry
     * @param opts The options for the request. Timeouts must be set in ms.
     * @returns The VAA bytes if available
     * @throws Errors if the VAA is not available after the retries
     */
    getVaaBytes<C extends Chain>(chain: C, emitter: AccountAddress<C>, sequence: bigint, timeout?: number): Promise<Uint8Array | undefined>;
    /**
     * Gets a VAA from the API or Guardian RPC, finality must be met before the VAA will be available.
     * @param chain The chain name
     * @param emitter The emitter address
     * @param sequence The sequence number
     * @param timeout The total amount of time to wait for the VAA to be available
     * @returns The VAA if available
     * @throws Errors if the VAA is not available after the retries
     */
    getVaa<T extends PayloadLiteral | PayloadDiscriminator, C extends Chain>(chain: C, emitter: AccountAddress<C>, sequence: bigint, decodeAs: T, timeout?: number): Promise<ReturnType<typeof deserialize<T>> | undefined>;
    getCircleAttestation(msgHash: string, timeout?: number): Promise<string | null>;
    /**
     * Get the status of a transaction, identified by the chain, emitter address, and sequence number
     *
     * @param chain the chain name
     * @param emitter the emitter address
     * @param sequence the sequence number
     * @returns the TransactionStatus
     */
    getTransactionStatus(chain: Chain, emitter: UniversalAddress | NativeAddress<Platform>, sequence: bigint, timeout?: number): Promise<TransactionStatus>;
    /**
     * Parse an address to a universal address
     *
     * @param address The native address
     * @returns The address in the universal format
     */
    parseAddress<C extends Chain>(chain: C, address: string): NativeAddress<C>;
    /**
     * Parses all relevant information from a transaction given the sending tx hash and sending chain
     *
     * @param tx The sending transaction hash
     * @param chain The sending chain name or id
     * @returns The parsed WormholeMessageId
     */
    parseMessageFromTx(chain: Chain, txid: TxHash, timeout?: number): Promise<WormholeMessageId[]>;
}
//# sourceMappingURL=wormhole.d.ts.map