import { encoding } from "@wormhole-foundation/sdk-base";
import { TokenBridge, isTransactionIdentifier, isWormholeMessageId, nativeChainAddress, toNative, } from "@wormhole-foundation/sdk-definitions";
import { signSendWait } from "../common";
import { isTokenTransferDetails } from "../types";
import { TransferState } from "../wormholeTransfer";
export class TokenTransfer {
    wh;
    // state machine tracker
    state;
    // transfer details
    transfer;
    // txids, populated once transactions are submitted
    txids = [];
    // The corresponding vaa representing the TokenTransfer
    // on the source chain (if its been completed and finalized)
    vaas;
    constructor(wh, transfer) {
        this.state = TransferState.Created;
        this.wh = wh;
        this.transfer = transfer;
    }
    async getTransferState() {
        if (!this.transfer.automatic)
            return this.state;
        if (!this.vaas || this.vaas.length === 0)
            return this.state;
        const { chain, emitter, sequence } = this.vaas[0].id;
        const txStatus = await this.wh.getTransactionStatus(chain, emitter, sequence);
        if (txStatus.globalTx.destinationTx) {
            switch (txStatus.globalTx.destinationTx.status) {
                case "completed":
                    this.state = TransferState.Completed;
                    break;
                // ... more?
            }
        }
        return this.state;
    }
    static async from(wh, from, timeout = 6000) {
        if (isTokenTransferDetails(from)) {
            await TokenTransfer.validateTransferDetails(wh, from);
            // Bit of (temporary) hackery until solana contracts support being
            // sent a VAA with the primary address
            if (from.to.chain === "Solana") {
                // source token id
                const token = from.token === "native"
                    ? await wh.getChain(from.from.chain).getNativeWrappedTokenId()
                    : from.token;
                from.to = await wh.getTokenAccount(from.to, token);
            }
            if (from.to.chain === "Sei") {
                // For sei, we reserve the payload for a token transfer through the sei bridge.
                from.payload = encoding.bytes.encode(JSON.stringify({
                    basic_recipient: {
                        recipient: encoding.b64.encode(from.to.address.toString()),
                    },
                }));
                from.to = nativeChainAddress(from.to.chain, wh.getContracts("Sei").translator);
            }
            return new TokenTransfer(wh, from);
        }
        let tt;
        if (isWormholeMessageId(from)) {
            tt = await TokenTransfer.fromIdentifier(wh, from, timeout);
        }
        else if (isTransactionIdentifier(from)) {
            tt = await TokenTransfer.fromTransaction(wh, from, timeout);
        }
        else {
            throw new Error("Invalid `from` parameter for TokenTransfer");
        }
        await tt.fetchAttestation(timeout);
        return tt;
    }
    // init from the seq id
    static async fromIdentifier(wh, id, timeout) {
        const vaa = await TokenTransfer.getTransferVaa(wh, id, timeout);
        const { chain, address } = vaa.payload.to;
        const { tokenBridgeRelayer } = wh.config.chains[chain].contracts;
        const relayerAddress = tokenBridgeRelayer
            ? nativeChainAddress(chain, tokenBridgeRelayer).address.toUniversalAddress()
            : null;
        // Check if its a payload 3 targeted at a relayer on the destination chain
        const automatic = vaa.payloadName === "TransferWithPayload" &&
            !!relayerAddress &&
            address.equals(relayerAddress);
        const token = vaa.payload.token;
        const details = {
            token: { chain: token.chain, address: token.address },
            amount: token.amount,
            // TODO: the `from.address` here is a lie, but we don't
            // immediately have enough info to get the _correct_ one
            from: { chain: id.chain, address: id.emitter },
            to: { ...vaa.payload.to },
            automatic,
        };
        const tt = new TokenTransfer(wh, details);
        tt.vaas = [{ id: id, vaa }];
        tt.state = TransferState.Attested;
        return tt;
    }
    static async fromTransaction(wh, from, timeout) {
        const msg = await TokenTransfer.getTransferMessage(wh, from, timeout);
        const tt = await TokenTransfer.fromIdentifier(wh, msg, timeout);
        tt.txids = [from];
        return tt;
    }
    // start the WormholeTransfer by submitting transactions to the source chain
    // returns a transaction hash
    async initiateTransfer(signer) {
        /*
            0) check that the current `state` is valid to call this (eg: state == Created)
            1) get a token transfer transaction for the token bridge given the context
            2) sign it given the signer
            3) submit it to chain
            4) return transaction id
        */
        if (this.state !== TransferState.Created)
            throw new Error("Invalid state transition in `start`");
        this.txids = await TokenTransfer.transfer(this.wh.getChain(this.transfer.from.chain), this.transfer, signer);
        this.state = TransferState.Initiated;
        return this.txids.map(({ txid }) => txid);
    }
    // wait for the VAA to be ready
    // returns the sequence number
    async fetchAttestation(timeout) {
        /*
            0) check that the current `state` is valid to call this  (eg: state == Started)
            1) poll the api on an interval to check if the VAA is available
            2) Once available, pull the VAA and parse it
            3) return seq
        */
        if (this.state < TransferState.Initiated || this.state > TransferState.Attested)
            throw new Error("Invalid state transition in `ready`");
        if (!this.vaas || this.vaas.length === 0) {
            if (this.txids.length === 0)
                throw new Error("No VAAs set and txids available to look them up");
            // TODO: assuming the _last_ transaction in the list will contain the msg id
            const txid = this.txids[this.txids.length - 1];
            const msgId = await TokenTransfer.getTransferMessage(this.wh, txid, timeout);
            this.vaas = [{ id: msgId }];
        }
        for (const idx in this.vaas) {
            // Check if we already have the VAA
            if (this.vaas[idx].vaa)
                continue;
            this.vaas[idx].vaa = await TokenTransfer.getTransferVaa(this.wh, this.vaas[idx].id, timeout);
        }
        this.state = TransferState.Attested;
        return this.vaas.map((v) => v.id);
    }
    // finish the WormholeTransfer by submitting transactions to the destination chain
    // returns a transaction hash
    async completeTransfer(signer) {
        /*
            0) check that the current `state` is valid to call this  (eg: state == Ready)
            1) prepare the transactions and sign them given the signer
            2) submit the VAA and transactions on chain
            3) return txid of submission
        */
        if (this.state < TransferState.Attested)
            throw new Error("Invalid state transition in `finish`. Be sure to call `fetchAttestation`.");
        if (!this.vaas)
            throw new Error("No VAA details available");
        // TODO: when do we get >1?
        const { vaa } = this.vaas[0];
        if (!vaa)
            throw new Error(`No VAA found for ${this.vaas[0].id.sequence}`);
        const toChain = this.wh.getChain(this.transfer.to.chain);
        const redeemTxids = await TokenTransfer.redeem(toChain, vaa, signer, this.transfer.automatic);
        this.txids.push(...redeemTxids);
        return redeemTxids.map(({ txid }) => txid);
    }
    // Static method to perform the transfer so a custom RPC may be used
    // Note: this assumes the transfer has already been validated with `validateTransfer`
    static async transfer(fromChain, transfer, signer) {
        const senderAddress = toNative(signer.chain(), signer.address());
        const tokenAddress = transfer.token === "native" ? "native" : transfer.token.address;
        let xfer;
        if (transfer.automatic) {
            const tb = await fromChain.getAutomaticTokenBridge();
            const fee = await tb.getRelayerFee(senderAddress, transfer.to, tokenAddress);
            xfer = tb.transfer(senderAddress, transfer.to, tokenAddress, transfer.amount, fee, transfer.nativeGas);
        }
        else {
            const tb = await fromChain.getTokenBridge();
            xfer = tb.transfer(senderAddress, transfer.to, tokenAddress, transfer.amount, transfer.payload);
        }
        return signSendWait(fromChain, xfer, signer);
    }
    // Static method to allow passing a custom RPC
    static async redeem(toChain, vaa, signer, automatic) {
        const signerAddress = toNative(signer.chain(), signer.address());
        let xfer;
        if (automatic) {
            if (vaa.payloadName === "Transfer")
                throw new Error("VAA is a simple transfer but expected Payload for automatic delivery");
            const tb = await toChain.getAutomaticTokenBridge();
            xfer = tb.redeem(signerAddress, vaa);
        }
        else {
            const tb = await toChain.getTokenBridge();
            xfer = tb.redeem(signerAddress, vaa);
        }
        return signSendWait(toChain, xfer, signer);
    }
    static async getTransferMessage(wh, tx, timeout) {
        const { chain, txid } = tx;
        const msgs = await wh.parseMessageFromTx(chain, txid, timeout);
        if (msgs.length === 0)
            throw new Error(`No messages found in transaction ${txid}`);
        return msgs[0];
    }
    static async getTransferVaa(wh, whm, timeout) {
        const { chain, emitter, sequence } = whm;
        const vaa = await wh.getVaa(chain, emitter, sequence, TokenBridge.getTransferDiscriminator(), timeout);
        if (!vaa)
            throw new Error(`No VAA available after retries exhausted`);
        return vaa;
    }
    static async validateTransferDetails(wh, transfer) {
        if (transfer.payload && transfer.automatic)
            throw new Error("Payload with automatic delivery is not supported");
        if (transfer.nativeGas && !transfer.automatic)
            throw new Error("Gas Dropoff is only supported for automatic transfers");
        const fromChain = wh.getChain(transfer.from.chain);
        const toChain = wh.getChain(transfer.to.chain);
        if (!fromChain.supportsTokenBridge())
            throw new Error(`Token Bridge not supported on ${transfer.from.chain}`);
        if (!toChain.supportsTokenBridge())
            throw new Error(`Token Bridge not supported on ${transfer.to.chain}`);
        if (transfer.automatic && !fromChain.supportsAutomaticTokenBridge())
            throw new Error(`Automatic Token Bridge not supported on ${transfer.from.chain}`);
        if (transfer.to.chain === "Sei" && transfer.payload)
            throw new Error("Arbitrary payloads unsupported for Sei");
        if (transfer.amount === 0n)
            throw new Error("Amount cannot be 0");
        // TODO: Make sure the token exists on the destination chain
        // const token: TokenId =
        //   transfer.token === "native" ? await fromChain.getNativeWrappedTokenId() : transfer.token;
        // await wh.getWrappedAsset(transfer.to.chain, token);
    }
}
//# sourceMappingURL=tokenTransfer.js.map